import os
import json
import time
import pickle
import sys

# import matplotlib.pyplot as plt

from binarycpython.utils.grid import Population
from binarycpython.utils.functions import (
    get_help_all,
    get_help,
    output_lines,
)

def test_setup_population():
    """
    Unit test for setting up the population object
    """

    test_pop = Population()

    assert isinstance(test_pop, Population), "Population object not created properly"

def test_set_value_population():
    """
    Unit test for setting values in the population object
    """

    test_pop = Population()

    test_pop.set(
        verbosity=1,
    )

    test_pop.set(
        M_1=10,
        data_dir='/tmp/binary-c/output',
        ensemble_filter_SUPERNOVAE=1,
    )

    assert test_pop.bse_options['M_1'] == 10, 'BSE options not correctly set'
    assert test_pop.grid_options['verbosity'] == 1, 'Grid options not correctly set'
    assert test_pop.custom_options['data_dir'] == '/tmp/binary-c/output', 'Custom options not correctly set'
    assert test_pop.bse_options['ensemble_filter_SUPERNOVAE'] == 1, "Parameters are not getting caught correctly (The wildcards of parameters with %d in their name)"

def test_set_argline_output_population():
    """
    Unit test for setting values in the population object
    """

    test_pop = Population()

    test_pop.set(
        M_1=10,
        M_2 = 14.0,  # Msun
        separation = 0,  # 0 = ignored, use period
        orbital_period = 4530.0,  # days
        eccentricity = 0.0,
        metallicity = 0.02,
        max_evolution_time = 15000,
        verbosity=1,
    )

    argline = test_pop.return_argline()
    assert argline == "binary_c M_1 10 M_2 14.0 eccentricity 0.0 max_evolution_time 15000 metallicity 0.02 orbital_period 4530.0 separation 0", "Argline not constructed correctly. Check if values are set properly."

def test_version_info_dict_population():
    """
    Unit test for setting values in the population object

    TODO: Add status to output of the version_info, with -1 if there are unmatched items
    """

    test_pop = Population()

    # return version info
    version_info = test_pop.return_binary_c_version_info(parsed=True)
    version_info_keys = version_info.keys()

    assert "L_SUN" in version_info_keys, "Missing item in version info"
    assert "ADAPTIVE_RLOF_LOG" in version_info_keys, "Missing item in version info"
    assert "GSL_VERSION" in version_info_keys, "Missing item in version info"
    assert "git_revision" in version_info_keys, "Missing item in version info"

def test_settings_output_population():
    """
    Unit test for outputting the settings of the population dict object
    """

    test_pop = Population()

    # return version info
    population_settings = test_pop.return_population_settings()
    population_settings_keys = population_settings.keys()

    assert 'bse_options' in population_settings_keys, "Missing information in the population_settings dict"
    assert 'grid_options' in population_settings_keys, "Missing information in the population_settings dict"
    assert 'custom_options' in population_settings_keys, "Missing information in the population_settings dict"

def test_all_info_population():
    """
    Unit test for outputting all information of the population object
    """

    test_pop = Population()

    # return all info:
    all_info_dict = test_pop.return_all_info()
    all_info_keys = all_info_dict.keys()

    assert 'population_settings' in all_info_keys, "Missing information in the all_info settings dict"
    assert 'binary_c_defaults' in all_info_keys, "Missing information in the all_info settings dict"
    assert 'binary_c_version_info' in all_info_keys, "Missing information in the all_info settings dict"
    assert 'binary_c_help_all' in all_info_keys, "Missing information in the all_info settings dict"

def test_evolve_single_system_population():
    """
    Unit test for the evolve_single_system function of the population object
    """

    test_pop = Population()

    test_pop.set(
        M_1=10,
        M_2 = 14.0,  # Msun
        separation = 0,  # 0 = ignored, use period
        orbital_period = 4530.0,  # days
        eccentricity = 0.0,
        metallicity = 0.02,
        max_evolution_time = 15000,
        verbosity=0,
    )

    output = test_pop.evolve_single()

    assert "SINGLE_STAR_LIFETIME" in output, "Failed to evolve a system"

# def test_custom_logging_memory_adress():

def test_C_auto_logging_population():
    """
    Unit test for the creating a custom logging output by setting the 
    C_auto_logging value with a list containing a headerline as key and a list of parameters as value
    """

    test_pop = Population()

    test_pop.set(
        M_1 = 14,
        M_2 = 10.0,  # Msun
        separation = 0,  # 0 = ignored, use period
        orbital_period = 4530.0,  # days
        eccentricity = 0.0,
        metallicity = 0.02,
        max_evolution_time = 15000,
        verbosity=0,
    )


    test_pop.set(C_auto_logging={'MY_HEADER_LINE': ['star[0].mass', 'star[1].mass', 'model.probability']})

    output = test_pop.evolve_single()

    print(output.splitlines()[0].split())

    first_line = output.splitlines()[0].split()

    assert first_line[0] == "MY_HEADER_LINE" , "Failed to set the custom logging correctly"
    assert first_line[1] == "14" , "Failed to set the custom logging correctly. First mass should be 14"
    assert first_line[2] == "10" , "Failed to set the custom logging correctly. Second mass should be 10"
    assert first_line[3] == "1" , "Failed to set the custom logging correctly. Probability should be 1"

def test_C_logging_code_population():
    """
    Unit test for the creating a custom logging output by setting the 
    C_logging_code which contains the exact print statement (including possible logic).
    This will be evaluated and put into binary_c
    """

    test_pop = Population()

    test_pop.set(
        M_1 = 14,
        M_2 = 10.0,  # Msun
        separation = 10000,  # 0 = ignored, use period
        # orbital_period = 4530.0,  # days
        eccentricity = 0.0,
        metallicity = 0.02,
        max_evolution_time = 15000,
        verbosity=0,
    )

    test_pop.set(C_logging_code='Printf("MY_STELLAR_DATA mass=%g separation=%g probability=%g\\n", stardata->star[0].mass, stardata->common.orbit.separation, stardata->model.probability);')

    output = test_pop.evolve_single()
    first_line = output.splitlines()[0].split()

    assert first_line[0] == "MY_STELLAR_DATA" , "Failed to set the custom logging correctly. Headerline should be MY_STELLAR_DATA"
    assert first_line[1] == "mass=14" , "Failed to set the custom logging correctly. First mass should be 14"
    assert first_line[2] == "separation=10000" , "Failed to set the custom logging correctly. Separation should be 10000"
    assert first_line[3] == "probability=1" , "Failed to set the custom logging correctly. Probability should be 1"

def parse_function(self, output):
    """
    Dummy parse function for handling the output of a system
    """

    parameters = ['mass_1', 'separation', 'probability']
    separator='\t'

    outfilename = os.path.join(self.grid_options['tmp_dir'], 'output_parse_function_test.txt')
    if os.path.isfile(outfilename):
        os.remove(outfilename)

    for el in output_lines(output):
        headerline = el.split()[0]

        if (headerline=='MY_STELLAR_DATA'):
            values = el.split()[1:]
            if not os.path.exists(outfilename):
                with open(outfilename, 'w') as f:
                    f.write(separator.join(parameters)+'\n')

            with open(outfilename, 'a') as f:
                f.write(separator.join(values)+'\n')

def test_parse_function_population():
    """
    Unit test to test the ise of a parsing function for the output
    """

    test_pop = Population()

    test_pop.set(
        M_1=10,
        M_2 = 14.0,  # Msun
        separation = 0,  # 0 = ignored, use period
        orbital_period = 4530.0,  # days
        eccentricity = 0.0,
        metallicity = 0.02,
        max_evolution_time = 15000,
        verbosity=0,
    )

    test_pop.set(
        C_logging_code='Printf("MY_STELLAR_DATA mass=%g separation=%g probability=%g\\n", stardata->star[0].mass, stardata->common.orbit.separation, stardata->model.probability);'
    )

    test_pop.set(
        parse_function=parse_function
    )

    # Run with parse function
    test_pop.evolve_single(parse_function)

    outfilename = os.path.join(test_pop.grid_options['tmp_dir'], 'output_parse_function_test.txt')
    assert os.path.isfile(outfilename), "Output file not created!"

    with open(outfilename, 'r') as outfile:
        output = outfile.readlines()
    first_line_split = output[0].strip().split("\t")

    assert first_line_split[0] == 'mass_1', "Output header not created correctly"
    assert first_line_split[1] == 'separation', "Output header not created correctly"
    assert first_line_split[2] == 'probability', "Output header not created correctly"
    assert len(output) > 1, "File doesn't seem to contain any real data"

if __name__ == "__main__":
    # test_setup_population()
    # test_set_value_population()
    # test_set_argline_output_population()
    # test_version_info_dict_population()
    # test_settings_output_population()
    # test_all_info_population()
    # test_evolve_single_system_population()
    # test_C_auto_logging_population()
    # test_C_logging_code_population()
    # test_parse_function_population()
    print('yo')

quit()

# print(len(test_pop.return_binary_c_defaults()))
# print('\n\n')
# print(len(test_pop.cleanup_defaults()))

## Use custom arg file
# test_pop.evolve_population(custom_arg_file='/home/david/projects/binary_c_root/binary_c-python/tests/population/custom_arg_file.txt')


### Grid generating testing
# test_pop.add_grid_variable(
#     name="lnm1",
#     longname="log primary mass",
#     valuerange=[10, 20],
#     resolution="10",
#     spacingfunc="np.linspace(0.213, 10.2, 10)",
#     precode="M_1=math.exp(lnm1)",
#     probdist="flat(M_1)",
#     # probdist='self.custom_options["extra_prob_function"](M_1)',
#     dphasevol="",
#     parameter_name="M_1",
#     condition="",
# )

# ### Grid generating test.
# test_pop.add_grid_variable(
#     name="period",
#     longname="period",
#     valuerange=["M_1", 20],
#     resolution="10",
#     spacingfunc="np.linspace(1, 10, 10)",
#     precode="orbital_period = period**2",
#     probdist="flat(orbital_period)",
#     parameter_name="orbital_period",
#     dphasevol="",
#     condition='self.grid_options["binary"]==0',
# )

# test_pop.generate_grid_code()

# test_pop.load_grid_function()

# test_pop.write_binary_c_calls_to_file(output_dir="/home/david/Desktop")


# print(test_pop.grid_options["code_string"])

####
# Dry run:
# test_pop.set(verbose=1)
# test_pop.add_grid_variable(
#     name="lnm1",
#     longname="log primary mass",
#     valuerange=[10, 20],
#     resolution="10",
#     spacingfunc="np.linspace(0.213, 10.2, 10)",
#     precode="M_1=math.exp(lnm1)",
#     probdist="flat(M_1)",
#     # probdist='self.custom_options["extra_prob_function"](M_1)',
#     dphasevol="",
#     parameter_name="M_1",
#     condition="",
# )

# ### Grid generating test.
# test_pop.add_grid_variable(
#     name="period",
#     longname="period",
#     valuerange=["M_1", 20],
#     resolution="10",
#     spacingfunc="np.linspace(1, 10, 10)",
#     precode="orbital_period = period**2",
#     probdist="flat(orbital_period)",
#     parameter_name="orbital_period",
#     dphasevol="",
#     condition='self.grid_options["binary"]==0',
# )

# test_pop.generate_grid_code(dry_run=True)

# test_pop.load_grid_function()

# test_pop.dry_run()

# ###
# testing population:
test_pop.set(
    verbose=1, amt_cores=2, binary=0,
)


resolution = {"M_1": 10, "per": 10}

test_pop.add_grid_variable(
    name="M_1",
    longname="log primary mass",
    valuerange=[10, 100],
    resolution="{}".format(resolution["M_1"]),
    spacingfunc="const(10, 100, {})".format(resolution["M_1"]),
    # precode="M_1=math.exp(lnm1)",
    probdist="flat(M_1)",
    # probdist='self.custom_options["extra_prob_function"](M_1)',
    dphasevol="dM_1",
    parameter_name="M_1",
)

### Grid generating test.
test_pop.add_grid_variable(
    name="period",
    longname="period",
    valuerange=["M_1", 20],
    resolution="{}".format(resolution["per"]),
    spacingfunc="np.linspace(1, 10, {})".format(resolution["per"]),
    precode="orbital_period = period**2",
    probdist="flat(orbital_period)",
    parameter_name="orbital_period",
    dphasevol="dper",
    condition='self.grid_options["binary"]==0',
)

test_pop.set(verbose=1, amt_cores=2, binary=0, evolution_type="linear")


# test_pop.evolve_population()
# test_pop.generate_grid_code()
# test_pop.load_grid_function()
# print(test_pop.grid_options['system_generator'])
# test_pop.grid_options['system_generator'](test_pop)
